import io
import pydyf
import shutil
import datetime
from typing import List


class WeasyForm:

    def __init__(self, pdf: pydyf.PDF):
        self.pdf = pdf
        self.catalog = pdf.catalog

    def get_pages(self):
        page_objects = list(filter(lambda _object: isinstance(_object, pydyf.Dictionary) and _object.get('Type') == '/Page', self.pdf.objects))
        pages = {}
        for index, page_object in enumerate(page_objects):
            pages[index] = page_object

        return pages

    def get_signatures(self):
        signature_objects = list(filter(lambda _object: isinstance(_object, pydyf.Dictionary) and _object.get('Type') == '/Sig', self.pdf.objects))
        signatures = {}
        for signature_object in signature_objects:
            signatures[signature_object.reference] = signature_object

        return signatures

    def find_page(self, page_number: int) -> pydyf.Dictionary:
        return self.get_pages()[page_number]

    def find_form_fields(self, field_type: str, field_title: str = None) -> List[pydyf.Dictionary]:
        def _form_field_filter(_object: pydyf.Object):

            if not isinstance(_object, pydyf.Dictionary):
                return False

            if _object.get('FT') != field_type:
                return False

            if field_title and pydyf.String(field_title).data != _object.get('T').data:
                return False

            return True

        return list(filter(_form_field_filter, self.pdf.objects))

    def register_annotation(self, page_object: pydyf.Dictionary, annotation_object: pydyf.Dictionary) -> None:
        found_annots = page_object.get('Annots')
        if found_annots:
            found_annots.extend(annotation_object.reference)
        else:
            annots = pydyf.Dictionary({
                'Annots': pydyf.Array([annotation_object.reference])
            })

            page_object.update(annots)

    def add_object(self, object_) -> None:
        self.pdf.add_object(object_)

    def append_empty_cryptographic_signature(self):
        signature_max_length = 16384 * 2
        date = datetime.datetime.utcnow()
        date = date.strftime('%Y%m%d%H%M%S+00\'00\'')
        form_signature = pydyf.Dictionary({
            'Type': '/Sig',
            'Filter': '/Adobe.PPKLite',
            'SubFilter': '/ETSI.CAdES.detached',
            'ByteRange[0 ********** ********** **********]': '/M(D:{date})'.format(date=date),
            #'Contents<{zeros}>'.format(zeros='0' * signature_max_length): ''
        })

        self.pdf.add_object(form_signature)

        return form_signature

    def write(self, target: str):
        file_obj = io.BytesIO()
        self.pdf.write(file_obj)
        if target is None:
            return file_obj.getvalue()
        else:
            file_obj.seek(0)
            if hasattr(target, 'write'):
                shutil.copyfileobj(file_obj, target)
            else:
                with open(target, 'wb') as fd:
                    shutil.copyfileobj(file_obj, fd)



from typing import Tuple, TYPE_CHECKING
import pydyf
from ..exceptions import SignatureException, PdfException

if TYPE_CHECKING:
    from ..WeasyForm import WeasyForm


class SignatureFormField(pydyf.Dictionary):
    def __init__(self, field_name: str, on_page: pydyf.Dictionary = None, box: pydyf.Array = None, annotation_flag: int = 132):
        rect = box if box else pydyf.Array([0, 0, 0, 0])

        data = {
            'FT': '/Sig',
            'T': pydyf.String(field_name),
            'Type': '/Annot',
            'Subtype': '/Widget',
            'F': annotation_flag,
            'Rect': rect,
        }

        if on_page:
            data['P'] = on_page.reference

        super(SignatureFormField, self).__init__(data)


class Signature:
    def __init__(self,
                 signer_pdf: 'WeasyForm',
                 signature_field_name: str,
                 on_page_number: int = 0,
                 signature_box: Tuple[int, int, int, int] = None,
                 ):
        self.signer_pdf = signer_pdf
        self.signature_field_name = signature_field_name
        self.signature_box = signature_box

        self.page_object = signer_pdf.find_page(on_page_number)

    def create_signature_field(self, signature_field_has_to_exists: bool = False) -> pydyf.Dictionary:
        acro_form = self.signer_pdf.catalog.get('AcroForm')
        if acro_form:
            fields = acro_form.get('Fields')
            if fields is None:
                raise PdfException('Mangled PDF: /Fields are missing in /AcroForm')
            form_field_signatures = [field for field in fields]

            found_suitable_signature_field = None
            for found_signature_field in self.signer_pdf.find_form_fields('/Sig', self.signature_field_name):
                # is registered in Fields?
                if found_signature_field.reference not in form_field_signatures:
                    continue

                # What is here is our sig field
                # Check if it is not already signed
                has_value = found_signature_field.get('V')
                if has_value:
                    raise SignatureException('Requested /Sig field with name {} is already filled.'.format(self.signature_field_name))

                found_suitable_signature_field = found_signature_field

            if not found_suitable_signature_field and signature_field_has_to_exists:
                raise SignatureException('Requested required /Sig field with name {} Was not found.'.format(self.signature_field_name))

            if found_suitable_signature_field:
                return found_suitable_signature_field
        else:
            # Create AcroForm
            fields = pydyf.Array([])
            acro_form = pydyf.Dictionary({
                'AcroForm': pydyf.Dictionary({
                    'Fields': fields
                })
            })

            self.signer_pdf.catalog.update(acro_form)

        signature_field = SignatureFormField(
            self.signature_field_name,
            self.page_object,
            pydyf.Array([i for i in self.signature_box])
        )

        self.signer_pdf.add_object(signature_field)

        fields.append(signature_field.reference)

        self.signer_pdf.register_annotation(self.page_object, signature_field)

        return signature_field

    def set_signature_flags(self, flag: int = 3):
        acro_form = self.signer_pdf.catalog.get('AcroForm')
        acro_form.update(pydyf.Dictionary({
            'SigFlags': flag
        }))

    def create_visible_signature_box(self, signature_field: pydyf.Dictionary, display: bool = True):
        llx, lly, urx, ury = self.signature_box
        w = abs(urx - llx)
        h = abs(ury - lly)
        if w and h:
            ap_dict = pydyf.Dictionary()
            if display:
                ap_stream = pydyf.Stream([], {
                    'BBox': pydyf.Array([0, h, w, 0]),
                    'Resources': pydyf.Dictionary(),
                    'Type': '/XObject',
                    'Subtype': '/Form',
                })
                ap_stream.push_state()

                # Background
                ap_stream.push_state()
                ap_stream.set_color_rgb('0.95', '0.95', '0.95')
                ap_stream.rectangle(0, 0, w, h)
                ap_stream.fill()
                ap_stream.pop_state()

                # Border
                ap_stream.set_line_width('0.5')
                ap_stream.rectangle(0, 0, w, h)
                ap_stream.stroke()
                ap_stream.pop_state()

            else:
                ap_stream = pydyf.Stream([], {
                    'BBox': pydyf.Array([0, h, w, 0]),
                    'Resources': pydyf.Dictionary(),
                    'Type': '/XObject',
                    'Subtype': '/Form',
                })

            self.signer_pdf.add_object(ap_stream)
            ap_dict['N'] = ap_stream.reference

            signature_field['AP'] = ap_dict

    @staticmethod
    def add_field(
            signer_pdf: 'WeasyForm',
            signature_field_name: str,
            on_page_number: int = 0,
            signature_box: Tuple[int, int, int, int] = None,
            is_signature_visible: bool = True,
            signature_field_has_to_exists: bool = False
    ) -> pydyf.Dictionary:
        signature = Signature(
            signer_pdf,
            signature_field_name=signature_field_name,
            on_page_number=on_page_number,
            signature_box=signature_box,
        )

        signature_field = signature.create_signature_field(signature_field_has_to_exists)
        signature.set_signature_flags()

        if signature_box:
            signature.create_visible_signature_box(signature_field, is_signature_visible)

        return signature_field




import pydyf

from .WeasyForm import WeasyForm
from weasyform import Document
from weasyprint.document import Matrix


from .fields.signature import Signature


class FormFinisher:

    def __init__(self, target: str = None, inject_empty_cryptographic_signature: bool = False):
        self.target = target
        self.inject_empty_cryptographic_signature = inject_empty_cryptographic_signature

    def __call__(self, pdf_document: Document, pdf: pydyf.PDF):
        weasy_form = WeasyForm(pdf)

        for page in pdf_document.pages:
            for form_element in page.form_elements:
                element_type, element_name, rectangle = form_element
                if element_type == 'signature':
                    zoom = 1
                    scale = zoom * 0.75
                    matrix = Matrix(scale, 0, 0, -scale, 0, page.height * scale)
                    x1, y1 = matrix.transform_point(*rectangle[:2])
                    x2, y2 = matrix.transform_point(*rectangle[2:])

                    signature = Signature.add_field(
                        weasy_form,
                        signature_field_name=element_name,
                        signature_box=(x1, y1, x2, y2)
                    )

                    if self.inject_empty_cryptographic_signature:
                        form_signature = weasy_form.append_empty_cryptographic_signature()

                        signature.update({
                            'V': form_signature.reference
                        })

        weasy_form.write(self.target)



import math
from weasyprint import HTML as WPHTML
from weasyprint.document import DocumentMetadata, Matrix, rectangle_aabb, Document as WPDocument, Page as WPPage
from weasyprint.text.fonts import FontConfiguration
from weasyprint.css.counters import CounterStyle
from weasyprint.formatting_structure.build import build_formatting_structure
from weasyprint.layout import layout_document
from weasyprint.layout.percent import percentage
from weasyprint.html import get_html_metadata
from weasyprint.formatting_structure import boxes


class Page(WPPage):
    form_elements = []

    def _gather_links_and_bookmarks(self, box, parent_matrix=None):

        # Get box transformation matrix.
        # "Transforms apply to block-level and atomic inline-level elements,
        #  but do not apply to elements which may be split into
        #  multiple inline-level boxes."
        # http://www.w3.org/TR/css3-2d-transforms/#introduction
        if box.style['transform'] and not isinstance(box, boxes.InlineBox):
            border_width = box.border_width()
            border_height = box.border_height()
            origin_x, origin_y = box.style['transform_origin']
            offset_x = percentage(origin_x, border_width)
            offset_y = percentage(origin_y, border_height)
            origin_x = box.border_box_x() + offset_x
            origin_y = box.border_box_y() + offset_y

            matrix = Matrix(e=origin_x, f=origin_y)
            for name, args in box.style['transform']:
                a, b, c, d, e, f = 1, 0, 0, 1, 0, 0
                if name == 'scale':
                    a, d = args
                elif name == 'rotate':
                    a = d = math.cos(args)
                    b = math.sin(args)
                    c = -b
                elif name == 'translate':
                    e = percentage(args[0], border_width)
                    f = percentage(args[1], border_height)
                elif name == 'skew':
                    b, c = math.tan(args[1]), math.tan(args[0])
                else:
                    assert name == 'matrix'
                    a, b, c, d, e, f = args
                matrix = Matrix(a, b, c, d, e, f) @ matrix
            box.transformation_matrix = (
                    Matrix(e=-origin_x, f=-origin_y) @ matrix)
            if parent_matrix:
                matrix = box.transformation_matrix @ parent_matrix
            else:
                matrix = box.transformation_matrix
        else:
            matrix = parent_matrix

        bookmark_label = box.bookmark_label
        if box.style['bookmark_level'] == 'none':
            bookmark_level = None
        else:
            bookmark_level = box.style['bookmark_level']
        state = box.style['bookmark_state']
        link = box.style['link']
        anchor_name = box.style['anchor']
        has_bookmark = bookmark_label and bookmark_level
        # 'link' is inherited but redundant on text boxes
        has_link = link and not isinstance(box, (boxes.TextBox, boxes.LineBox))
        # In case of duplicate IDs, only the first is an anchor.
        has_anchor = anchor_name and anchor_name not in self.anchors

        has_form_element = box.element_tag in ['input']

        if has_bookmark or has_link or has_anchor or has_form_element:
            pos_x, pos_y, width, height = box.hit_area()
            if has_link:
                token_type, link = link
                assert token_type == 'url'
                link_type, target = link
                assert isinstance(target, str)
                if link_type == 'external' and box.is_attachment:
                    link_type = 'attachment'
                if matrix:
                    link = (
                        link_type, target,
                        rectangle_aabb(matrix, pos_x, pos_y, width, height),
                        box.download_name)
                else:
                    link = (
                        link_type, target,
                        (pos_x, pos_y, pos_x + width, pos_y + height),
                        box.download_name)
                self.links.append(link)
            if matrix and (has_bookmark or has_anchor):
                pos_x, pos_y = matrix.transform_point(pos_x, pos_y)
            if has_bookmark:
                self.bookmarks.append(
                    (bookmark_level, bookmark_label, (pos_x, pos_y), state))
            if has_anchor:
                self.anchors[anchor_name] = pos_x, pos_y

            if has_form_element:
                element_type = box.element.attrib.get('type')
                name = box.element.attrib.get('name')
                self.form_elements.append((
                    element_type,
                    name,
                    rectangle_aabb(matrix, pos_x, pos_y, width, height) if matrix else (pos_x, pos_y, pos_x + width, pos_y + height)
                ))

        for child in box.all_children():
            self._gather_links_and_bookmarks(child, matrix)


class Document(WPDocument):
    @classmethod
    def _render(cls, html, stylesheets, presentational_hints=False,
                optimize_size=('fonts',), font_config=None, counter_style=None,
                image_cache=None):
        if font_config is None:
            font_config = FontConfiguration()

        if counter_style is None:
            counter_style = CounterStyle()

        context = cls._build_layout_context(
            html, stylesheets, presentational_hints, optimize_size,
            font_config, counter_style, image_cache)

        root_box = build_formatting_structure(
            html.etree_element, context.style_for, context.get_image_from_uri,
            html.base_url, context.target_collector, counter_style,
            context.footnotes)

        page_boxes = layout_document(html, root_box, context)
        rendering = cls(
            [Page(page_box) for page_box in page_boxes],
            DocumentMetadata(**get_html_metadata(html)),
            html.url_fetcher, font_config, optimize_size)
        return rendering


class HTML(WPHTML):
    def render(self, stylesheets=None, presentational_hints=False,
               optimize_size=('fonts',), font_config=None, counter_style=None,
               image_cache=None):

        return Document._render(
            self, stylesheets, presentational_hints,
            optimize_size, font_config, counter_style, image_cache)



from setuptools import setup, find_packages


def read_readme() -> str:
    with open('README.md', 'r', encoding='utf-8') as f:
        return f.read()


setup(
    name='weasyform',
    version='0.0.1',
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=[
        'WeasyPrint'
    ],

    url='https://github.com/Salamek/weasyform',
    license='BSD',
    author='Adam Schubert',
    author_email='adam.schubert@sg1-game.net',
    description='Simple form support for WeasyPrint',
    long_description=read_readme(),
    long_description_content_type='text/markdown',
    test_suite='tests',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development',
    ],
    python_requires='>=3.4',
    project_urls={
        'Release notes': 'https://github.com/Salamek/weasyform/releases',
    },
)



from weasyform.FormFinisher import FormFinisher
from weasyform import HTML

html = """
<style>
table {
    width: 100%;
    
}

table, th, td {
  border: 1px solid;
}

input {
    width: 100%;
    height: 20px;
}

</style>
<h1>TEST</h1>
<table>
<tr>
    <th>URL</th><td><a href="https://salamek.cz">This is a URL</a></td>
</tr>
<tr>
    <th>URL</th><td><textarea name="asdasd">dfsfsf</textarea></td>
</tr>
<tr>
    <th>Signature</th><td><input type="signature" name="testSignature" style="width: 100% height: 20px"></td>
</tr>
<tr>
    <th>Input text</th><td><input name="testText" type="text"></td>
</tr>
<tr>
    <th>Input number</th><td><input name="testNumber" type="number"></td>
</tr>
<tr>
    <th>Input email</th><td><input name="testEmail" type="email"></td>
</tr>
<tr>
    <th>Select</th><td><select name="testSelect"><option>A</option><option>B</option></select></td>
</tr>
</table>
"""

pdf_html = HTML(string=html).render()


target = 'forms.pdf'
pdf_html.write_pdf(finisher=FormFinisher(target=target))


